/* Initial beliefs and rules */

timeout(4000). 
part_num(50).
prev_cfp(27).
prev_propose(2).

/****************************************************************/
/****************************************************************/
/********** 				ICNP					*********/
/****************************************************************/
/****************************************************************/


/****************************************************************/
/********** 				PROTOCOL					*********/
/****************************************************************/

/* ICNP protocol, involves initiator and participant */
global_type(icnp, T) :- // FIPA ICNP
   part_num(PartNum) & 
   T = attrType(icnp, allDiff(_DL, _Init, PartNum, [], _), _).

attrType(icnp, allDiff(DL, Init, Current, List, X), 
fork(
  seq(sa(cfp(Init, X, DL),0),
  choice([attrType(ontime, dec_inc(X, Init, PrevCFP, _, PrevPropose, _, DL), _), attrType(late, dl(DL), _)])
),
attrType(icnp, allDiff(DL, Init, NewCurrent, [X|List], _), _))
) :- prev_cfp(PrevCFP) & prev_propose(PrevPropose) & Current > 0 & NewCurrent = Current-1.

attrType(icnp, allDiff(_DL, _Init, 0, _List, _X), lambda).

attrType(ontime, dec_inc(X, Init, _CFPprev, CFPcurr, _PROPprev, PROPcurr, DL), 
choice([seq(sa(ontime_refuse(X, Init, DL),0),lambda), 
 seq(sa(ontime_propose(X, Init, PROPcurr, DL),0),
   choice([
    choice([
     seq(sa(reject_proposal(Init, X),0),lambda),
     seq(sa(accept_proposal(Init, X),0), 
        choice([
           seq(sa(inform(X, Init),0),lambda), 
           seq(sa(failure(X, Init),0),lambda)
              ]))]),
    seq(sa(cfp(Init, X, CFPcurr),0),attrType(ontime, dec_inc(X, Init, CFPcurr, _, PROPcurr, _, noDl), _))
    ])
   )
  ])
 ).

attrType(late, dl(DL), 
choice([seq(sa(late_refuse(DL),0),lambda),seq(sa(late_propose(DL),0),lambda)])).
   
/****************************************************************/
/********** 		SENDING ACTION TYPES				*********/
/****************************************************************/

/* Types for the ICNP protocol */	  
role(AgName, participant) :- 
.term2string(AgName, NameStr) &
.substring("participant",NameStr, 0).

role(AgName, initiator) :- 
.term2string(AgName, "initiator").

has_type(msg(Init, X, tell, m(cfp(DL), _Cid)), cfp(Init, X, DL)) :- role(Init, initiator) & role(X, participant).

has_type(msg(Init, X, tell, m(cfp(CFPcurr, DL), _Cid)), cfp(Init, X, CFPcurr)) :-  role(Init, initiator) & role(X, participant). 

has_type(msg(X, Init, tell, m(propose(PROPcurr, noDl), _Cid)), ontime_propose(X, Init, PROPcurr, noDl)) :- role(Init, initiator) & role(X, participant).

has_type(msg(X, Init, tell, m(propose(PROPcurr, CurrentTime, DL), _Cid)), ontime_propose(X, Init, PROPcurr, DL)):- role(Init, initiator) & role(X, participant) & CurrentTime <= DL. 

has_type(msg(X, Init, tell, m(propose(PROPcurr, CurrentTime, DL), _Cid)), late_propose(DL)) :- .println(msg(X, Init, tell, m(propose(PROPcurr, CurrentTime, DL), _Cid))) & role(Init, initiator) & role(X, participant) & CurrentTime > DL.

has_type(msg(X, Init, tell, m(refuse(CurrentTime, DL), _Cid)), ontime_refuse(X, Init, DL)) :- role(Init, initiator) & role(X, participant) & CurrentTime <= DL. 

has_type(msg(X, Init, tell, m(refuse(CurrentTime, DL), _Cid)), late_refuse(DL)) :- role(Init, initiator) & role(X, participant) & CurrentTime > DL. 

has_type(msg(Init, X, tell, m(reject_proposal, _Cid)), reject_proposal(Init, X)) :- role(Init, initiator) & role(X, participant). 

has_type(msg(Init, X, tell, m(accept_proposal, _Cid)), accept_proposal(Init, X)) :- role(Init, initiator) & role(X, participant). 

has_type(msg(X, Init, tell, m(inform, _Cid)), inform(X, Init)) :- role(Init, initiator) & role(X, participant). 

has_type(msg(X, Init, tell, m(failure, _Cid)), failure(X, Init)) :- role(Init, initiator) & role(X, participant). 

 
dl(_).
dec_inc(_X, _Init, Cp, Cc, Pp, Pc, _) :- Cc = Cp-1 & Pc = Pp+1 & Cc > Pc.
allDiff(_DL, Init, _Num, [], X).
allDiff(_DL, Init, _Num, [H|T], X) :-  not(.member(X, [H|T])).


/****************************************************************/
/********** 				NEXT PREDICATE				*********/
/****************************************************************/

next(0,seq(sa(AType,N),T),AMessage,T,N) :- has_type(AMessage, AType).
next(M,seq(sa(AType),T1),AMessage,T1,N) :- has_type(AMessage, AType) & M > 0 & N = M - 1.
next(M,choice([T1,_]),A,T2,N) :- next(M,T1,A,T2,N). 
next(M,choice([_,T1]),A,T2,N) :- next(M,T1,A,T2,N).
next(N,fork(T1,T2),A,fork(T3,T2),M) :- next(N,T1,A,T3,M). 
next(N,fork(T1,T2),A,fork(T1,T3),M) :- next(N,T2,A,T3,M).
next(N,fork(T1,T2),A,fork(T3,T4),P) :- next(N,T1,A,T3,M) & M > 0 & next(M,T2,A,T4,P). 
next(N,fork(T1,T2),A,fork(T4,T3),P) :- next(N,T2,A,T3,M) & M > 0 & next(M,T1,A,T4,P). 
next(M,cat(T1,T2),A,cat(T3,T2),N) :- next(M,T1,A,T3,N).
next(M,cat(T1,T2),A,T3,N) :- empty(T1) & next(M,T2,A,T3,N).
next(M, attrType(Name, Constr, T1), A, T2, N) :- 
attrType(Name, Constr, T1) &
next(M, T1, A, T2, N) &
Constr. 

/****************************************************************/
/********** 				EMPTY PREDICATE				*********/
/****************************************************************/

empty(lambda).
empty(choice([T1,_])) :- empty(T1).
empty(choice([_,T1])) :- empty(T1).
empty(fork(T1,T2)) :- empty(T1) & empty(T2).
empty(cat(T1,T2)) :- empty(T1) & empty(T2).


/****************************************************************/
/********** 			FINITE CONSTRUCTOR		*********/
/****************************************************************/


finite_fork(T, T1, N) :- N>0 & gtu.copy_term(T1, Fresh1) & gtu.copy_term(T1, Fresh2) & N1 = N-1 & finite_fork(T2, Fresh1, N1) &  T = fork(Fresh2,T2).
finite_fork(lambda, _, 0).

/****************************************************************/
/********** 			TYPE_CHECK PREDICATE			*********/
/****************************************************************/

type_check(msg(S, R, P, C), NewState) :- 
   current_state(LastState, CID) &
   // .print( "\n", msg(S, R, P, C), "\n") & 
   // .print("inside type_check: LastState = ", LastState , "\n") & 
   next(0, LastState, msg(S, R, P, C), NewState, 0).


/****************************************************************/
/********** 				CURRENT STATE				*********/
/****************************************************************/   
   
current_state(initial).
current_id(0).

/****************************************************************/
/********** 				INITIAL GOAL				*********/
/****************************************************************/

/* Plans */

/* Monitor's plans */
   
@conversStartReq[atomic]
+start_conversation(Name)[source(S)]: 
current_id(Id) & NewId = Id+1 & global_type(Name, G) & CId = conv(S, Id)
   <-  -starting_time(_); 
       -start_conversation(Name)[source(S)]; .print("\nMessage ", start_conversation(Name), " from source ", S, " causes the start of the conversation\nThe global type is ", G); -current_id(Id); 
       +current_id(NewId);  
	   .send(S, tell, conversation_id(Name, CId)); 
	   +ongoing_conversation(Name, CId, S);
	   +current_state(initial, CId); 
	   gtu.get_current_ms(CurrentTime); 
	   +starting_time(CurrentTime);
	   !move_to_state(G, CId).   

@conversStopReq1[atomic]
+stop_conversation(CId)[source(S)]: 
ongoing_conversation(Name, CId, S) & current_state(LastState, CId) & empty(LastState)
   <-  -stop_conversation(CId)[source(S)]; 
       -ongoing_conversation(Name, CId, S);
	   +closed_conversation(Name, CId, S);
	   -current_state(LastState, CId);
       .print("\nMessage ", stop_conversation(CId), " from source ", S, " causes the succesful termination of the conversation\n").  	   

@conversStopReq2[atomic]
+stop_conversation(CId)[source(S1)]: 
ongoing_conversation(Name, CId, S2) & not(S1=S2)
   <-  .print("\nMessage ", stop_conversation(CId), " from source ", S1, " cannot stop a conversation started by ", S2, ": request ignored\n"). 

@conversStopReq3[atomic]
+stop_conversation(CId)[source(S)]: 
ongoing_conversation(Name, CId, S) & current_state(LastState, CId) & not(empty(LastState))
   <-  .print("\nMessage ", stop_conversation(CId), " from source ", S, " in state ", LastState, " cannot be accepted since the state is not empty: request ignored\n").   
   
@move2state[atomic]
+!move_to_state(NewState, CId) : current_state(LastState, CId) 
   <- -current_state(_, CId); 
      +current_state(NewState, CId). 

@successfulMove[atomic]
+!type_check_message(msg(S, R, P, m(C, CId))) : current_state(Current, CId) & ongoing_conversation(Name, CId, _) & type_check(msg(S, R, P, m(C, CId)), NewState)  
   <- !move_to_state(NewState, CId);
      .print("\nMessage ", msg(S, R, P, m(C, CId)), "\nleads from state ", Current, " to state ", NewState, "\n");
      .send(S, tell, ok_check(msg(S, R, P, m(C, CId)))).

@failingMoveAndProtocol
+!type_check_message(msg(S, R, P, m(C, CId))) : current_state(Current, CId) 
                                        & ongoing_conversation(Name, CId, Source)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, m(C, CId)), " received within protocol ", Name, "\ncannot be accepted in the current state ", Current, "\n"); -ongoing_conversation(Name, CId, Source); -current_state(Current, CId); +failed_conversation(Name, CId, Source).
	   
@messageReceptionOK
+msg(S, R, P, m(C, CId))[source(S)]: ongoing_conversation(_Name, CId, _Source) & timeout(MS)
   <- -msg(S, R, P, m(C, CId))[source(S)]; 
       // .print("\nReceived ", msg(S, R, P, m(C, CId))[source(S)], "\n");
      !type_check_message(msg(S, R, P, m(C, CId)));
	  !check_progress.   
   
@messageReceptionKO1
+msg(S, R, P, m(C, CId)): failed_conversation(Name, CId, Source)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, m(C, CId)), " received but conversation ", CId, " started by ", Source, " within protocol ", Name, " previously failed\n").

@messageReceptionKO2
+msg(S, R, P, m(C, CId)): closed_conversation(Name, CId, Source)
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, m(C, CId)), " received but conversation ", CId, " started by ", Source, " within protocol ", Name, " already closed successfully\n").
   
@messageReceptionKO3
+msg(S, R, P, m(C, CId)): not(failed_conversation(_, CId, _)) &               
                       not(ongoing_conversation(_, CId, _)) &
					   not(closed_conversation(_, CId, _))
   <- .print("\n*** DYNAMIC TYPE-CHECKING ERROR ***\nMessage ", msg(S, R, P, m(C, CId)), " received but conversation ", CId, " never started\n").
   
+!check_progress : timeout(MS) 
   <- .wait({+msg(S1, R1, P1, C1)}, MS, Event); 
	  !aux_check_progress(Event).
	  
+!aux_check_progress(Event) : timeout(MS) & Event < MS. 

+!aux_check_progress(Event) : timeout(MS) & Event >= MS  & starting_time(ST) & gtu.get_current_ms(CT) & T = (CT - ST) - 4000
   <- .print("\n*** WARNING ***\nNo progress for ", Event, " milliseconds\n"); 
   .print("\n*** Total execution time ***\n ", T). 
   // ;!check_progress. 
  
/****************************************************************/
/********** 				NEXT PREDICATE DEBUG VERSION	*********/
/****************************************************************/

/*

next(0,seq(sa(AType,N),T),AMessage,T,N) :- has_type(AMessage, AType) & .print("\nNEXT1\n", next(0,seq(sa(AType,N),T),AMessage,T,N)) & .print("\nNEXT1 AFTER\n").

next(M,seq(sa(AType),T1),AMessage,T1,N) :- .print("\nNEXT2, before has type\n", next(M,seq(sa(AType),T1),AMessage,T1,N)) & has_type(AMessage, AType) & .print("\nNEXT2 after has type\n", next(M,seq(sa(AType),T1),AMessage,T1,N)) & M > 0 & N = M - 1 & .print("\nNEXT2 AFTER\n").

next(M,choice([T1,_]),A,T2,N) :- .print("\nNEXT3\n", next(M,choice([T1,_]),A,T2,N)) & next(M,T1,A,T2,N) & .print("\nNEXT3 AFTER\n").

next(M,choice([_,T1]),A,T2,N) :- .print("\nNEXT4\n", next(M,choice([_,T1]),A,T2,N)) & next(M,T1,A,T2,N) & .print("\nNEXT4 AFTER\n").

next(N,fork(T1,T2),A,fork(T3,T2),M) :- .print("\nNEXT5\n", next(N,fork(T1,T2),A,fork(T3,T2),M)) & next(N,T1,A,T3,M) & .print("\nNEXT5 AFTER\n"). 

next(N,fork(T1,T2),A,fork(T1,T3),M) :- .print("\nNEXT6\n", next(N,fork(T1,T2),A,fork(T1,T3),M)) & next(N,T2,A,T3,M) & .print("\nNEXT6 AFTER\n").

next(N,fork(T1,T2),A,fork(T3,T4),P) :- .print("\nNEXT7\n", next(N,fork(T1,T2),A,fork(T3,T4),P)) & next(N,T1,A,T3,M) & M > 0 & next(M,T2,A,T4,P) & .print("\nNEXT7 AFTER\n"). 

next(N,fork(T1,T2),A,fork(T4,T3),P) :- .print("\nNEXT8\n", next(N,fork(T1,T2),A,fork(T4,T3),P)) & next(N,T2,A,T3,M) & M > 0 & next(M,T1,A,T4,P) & .print("\nNEXT8 AFTER\n"). 

next(M,cat(T1,T2),A,cat(T3,T2),N) :- next(M,T1,A,T3,N).
next(M,cat(T1,T2),A,T3,N) :- empty(T1) & next(M,T2,A,T3,N).

next(M, attrType(Name, Constr, T1), A, T2, N) :- 
.print("\nNEXT9\n", next(M, attrType(Name, Constr, T1), A, T2, N)) & attrType(Name, Constr, T1) &
next(M, T1, A, T2, N) & Constr & .print("\nNEXT9 AFTER\n"). 

*/

/* Plans for runtime type checking */

{include("./_session_tools.asl")}   
